# Standard library imports.
import datetime

# Pytz package imports.
import pytz

# Django package imports.
from django import forms
from django.forms import ValidationError
from django.utils.translation import ugettext as _

# Local package imports.
from .managers import FIVEMIN_FREQUENCY_ID
from .managers import HOURLY_FREQUENCY_ID
from .managers import DAILY_FREQUENCY_ID
from .managers import now


# These are the frequency ID values used to specify which table will 
# be used to retrieve data.
FREQUENCY_CHOICES = ((DAILY_FREQUENCY_ID, 'Daily'), 
    (HOURLY_FREQUENCY_ID, 'Hourly'),
    (FIVEMIN_FREQUENCY_ID, 'Five Minute'),)

# Data output constants.  These constants are used in the request and 
# validation code for viewing data.
GRAPHICAL_OUTPUT = 'graphical'
TABLE_OUTPUT = 'table'
CSV_OUTPUT = 'csv'
OUTPUT_CHOICES = ((GRAPHICAL_OUTPUT, 'Graphical Plots'), 
    (TABLE_OUTPUT, 'HTML Table'),
    (CSV_OUTPUT, 'CSV Data'),)

# Data units constants.  These constants are used in the request code 
# for viewing data and provide the ability to choose whether data will 
# be viewed in imperial or metric units.
IMPERIAL_UNITS = 'imperial'
METRIC_UNITS = 'metric'
UNITS_CHOICES = ((IMPERIAL_UNITS, 'English'), 
    (METRIC_UNITS, 'Metric'))


class StationTimeseries(forms.Form):

    station_id = forms.CharField()
    frequency_id = forms.ChoiceField(choices = FREQUENCY_CHOICES, 
        initial = DAILY_FREQUENCY_ID)
    start_date = forms.DateField(input_formats = ['%Y-%m-%d'])
    end_date = forms.DateField(input_formats = ['%Y-%m-%d'])
    time_zone = forms.CharField()
    units = forms.ChoiceField(choices = UNITS_CHOICES, 
        initial = IMPERIAL_UNITS)
    output = forms.ChoiceField(choices = OUTPUT_CHOICES, 
        initial = GRAPHICAL_OUTPUT)

    def clean_time_zone(self):

        time_zone = self.cleaned_data.get('time_zone')
        time_zone = time_zone.replace('--SLASH--', '/')

        if time_zone not in pytz.all_timezones:

            msg = _('The given time zone does not exist.')
            raise forms.ValidationError(msg, code = 'invalid')

        else:

            url_time_zone = time_zone.replace('/', '--SLASH--')
            self.cleaned_data['url_time_zone'] = url_time_zone

        return time_zone

    def clean_units(self):

        units = self.cleaned_data.get('units')

        if units == IMPERIAL_UNITS:
            self.cleaned_data['imperial'] = True

        else:
            self.cleaned_data['imperial'] = False

        return units

    def clean(self):

        if len(self.errors) == 0:

            frequency_id = self.cleaned_data.get('frequency_id')
            sd = self.cleaned_data.get('start_date')
            ed = self.cleaned_data.get('end_date')
            time_zone = self.cleaned_data.get('time_zone')
            output = self.cleaned_data.get('output')

            # Check if the start date is greater than the end date.  Raise 
            # a 'ValidationError' if True.
            if sd > ed:

                msg = _('Start date is greater than the end date.')
                raise forms.ValidationError(msg, code = 'invalid')

            # The delta between the start and end dates.  Used for 
            # checking against the output variable.
            delta = ed - sd

            # The validation error template.
            msg = _('%(output)s output for %(frequency)s data is not ' + \
                'allowed for date ranges greater than %(date_range)s!')
            _p = {}
            _c = 'invalid'

            # Determine the start and end datetimes and localize both 
            # datetimes using the 'time_zone' variable.
            sdt = datetime.datetime(sd.year, sd.month, sd.day)
            edt = datetime.datetime(ed.year, ed.month, ed.day)

            tzinfo = pytz.timezone(time_zone)
            sdt = tzinfo.localize(sdt)
            edt = tzinfo.localize(edt)

            if frequency_id == FIVEMIN_FREQUENCY_ID:

                _p['frequency'] = 'five minute'

                if output == GRAPHICAL_OUTPUT:

                    _p['output'] = 'Graphical'

                    if delta > datetime.timedelta(days = 7):

                        _p['date_range'] = 'one week'
                        raise ValidationError(msg, params = _p, code = _c)

                else:

                    _p['output'] = 'Table or CSV'

                    if delta > datetime.timedelta(days = 31):

                        _p['date_range'] = 'one month'
                        raise ValidationError(msg, params = _p, code = _c)

                sdt += datetime.timedelta(seconds = 300)
                edt += datetime.timedelta(days = 1)

            elif frequency_id == HOURLY_FREQUENCY_ID:

                _p['frequency'] = 'hourly'

                if output == GRAPHICAL_OUTPUT:

                    _p['output'] = 'Graphical'

                    if delta > datetime.timedelta(days = 31):

                        _p['date_range'] = 'one month'
                        raise ValidationError(msg, params = _p, code = _c)

                else:

                    _p['output'] = 'Table or CSV'

                    if delta > datetime.timedelta(days = 366):

                        _p['date_range'] = 'one year'
                        raise ValidationError(msg, params = _p, code = _c)

                sdt += datetime.timedelta(seconds = 3600)
                edt += datetime.timedelta(days = 1)

            else:

                _p['frequency'] = 'daily'

                if output == GRAPHICAL_OUTPUT:

                    _p['output'] = 'Graphical'

                    if delta > datetime.timedelta(days = 366):

                        _p['date_range'] = 'one year'
                        raise ValidationError(msg, params = _p, code = _c)

                else:

                    _p['output'] = 'Table or CSV'

                    if delta > datetime.timedelta(days = 1830):

                        _p['date_range'] = 'five years'
                        raise ValidationError(msg, params = _p, code = _c)

            self.cleaned_data['sdt'] = sdt
            self.cleaned_data['edt'] = edt

        return None


class StationTimeseriesPrev(StationTimeseries):

    start_date = None
    end_date = None
    prev_days = forms.IntegerField(min_value = 1, max_value = 8)

    def clean(self):

        if len(self.errors) == 0:

            prev_days = self.cleaned_data.get('prev_days')
            time_zone = self.cleaned_data.get('time_zone')

            # Get the current UTC datetime and convert it to the 
            # 'local' time zone provided by the 'time_zone' value.
            ltz = pytz.timezone(time_zone)
            ltznow = now().astimezone(ltz)

            # Create the start and end dates, using the current 
            # localized datetime and the 'previous_days' value.
            ed = ltznow.date()
            sd = ed - datetime.timedelta(days = (prev_days - 1))

            # Add the start and end dates to the 'cleaned_data' 
            # attribute.
            self.cleaned_data['start_date'] = sd
            self.cleaned_data['end_date'] = ed

            # Call the 'clean' method for the 'StationTimeseries' 
            # form class.
            StationTimeseries.clean(self)

        return None


class StationDailyRGTimeseries(StationTimeseries):

    frequency_id = forms.ChoiceField(
        choices = ((DAILY_FREQUENCY_ID, 'Daily'),), 
        initial = DAILY_FREQUENCY_ID)
    output = forms.ChoiceField(
        choices = ((CSV_OUTPUT, 'CSV'), 
            (TABLE_OUTPUT, 'HTML Table')), 
        initial = TABLE_OUTPUT)
    gdd_tbase = forms.FloatField(initial = 50.0,
        label = 'Base Temperature')
    gdd_tmaxcutoff = forms.FloatField(initial = 999.0, 
        label = 'Max Cutoff Temperature')
    gdd_tmincutoff = forms.FloatField(initial = -999.0, 
        label = 'Min Cutoff Temperature')
    cgdd_offset = forms.FloatField(initial = 0.0, 
        label = 'Cumulative GDD Offset')


class StationDailyRGTimeseriesPrev(StationDailyRGTimeseries, 
    StationTimeseriesPrev):

    start_date = None
    end_date = None

    def clean(self):

        if len(self.errors) == 0:

            StationTimeseriesPrev.clean(self)

        return None


