# Standard library imports.
import csv
import datetime
import re
import urllib2

# Pytz package imports.
import pytz

# Pandas package imports.
import pandas as pd

# Django package imports.
from django.core.management.base import BaseCommand
from django.core.management.base import CommandError

# Local package imports.
from ...models import now
from ...models import Station
from ...models import FiveMinData
from ...models import HourlyData


class Command(BaseCommand):

    def __parse_ids(self, url, **kwargs):

        try:

            f = urllib2.urlopen(url)
            reader = csv.reader(f, delimiter = ',', quoting = csv.QUOTE_MINIMAL)

            # Get the first row (line) only.
            row_one = reader.next()

            if row_one[0] == 'TOA5':

                station_id = re.search(r'\((.*?)\)', row_one[1]).group(1)\
                    .lower()
                frequency_id = row_one[7]

                if frequency_id not in ('fivemin', 'hourly',):
                    frequency_id = None

                if frequency_id is not None:
                    ids = (station_id, frequency_id,)

            else:
                ids = None

            f.close()

        except Exception as e:

            print e
            ids = None

        return ids

    def __parse_row(self, dt, row, bulk_create, update):

        station = None

        for _s in self.stations:
            if station == None:
                if _s.valid_dt(dt):
                    station = _s

        if station is not None:

            ed = self.existing_data.get(str(dt), None)
            data = None

            if ed is not None:
                if update is True:
                    data = ed

            else:
                data = self.model(dt = dt, station = station)

            if data is not None:

                for _v in self.toa5.columns:

                    if _v == 'RECORD':

                        try:
                            setattr(data, 'record', int(row[_v]))

                        except:
                            pass

                    else:

                        try:

                            value = float(row[_v])

                            if str(value) != str(float('nan')):
                                setattr(data, _v, value)

                        except:
                            pass

                if ed is not None:
                    if update is True:
                        data.save()

                else:

                    if bulk_create is True:
                        self.new_data[str(data.dt)] = data

                    else:
                        data.save()

        return None

    def __split(self, data_list):

        new_data_lists = []

        data_list.sort(key = lambda x: x.dt)
        data_list.reverse()

        # We get the count of the new data rows so we can decide how 
        # to split up the data in the __new_rows attribute
        num_data = len(data_list)

        threshold = 200
        interval = 100

        if num_data < threshold:
            new_data_lists.append(data_list)

        elif (num_data >= threshold):

            # Slice off the end of the new data list and reverse 
            # it so that the earliest data are saved first.
            earliest_data = data_list[num_data - interval:num_data]
            earliest_data.reverse()
            new_data_lists.append(earliest_data)

            # Slice off the beginning of the new data list list so 
            # that the latest data is saved first.
            latest_data = data_list[:interval]
            new_data_lists.append(latest_data)

            # Get the middle of the new data list and determine how 
            # much it needs to be split.
            middle_data = data_list[interval: num_data - interval]
            num_middle_data = len(middle_data)

            count = 0

            while count < num_middle_data:

                if count + interval >= num_middle_data:
                    sub_middle_data = \
                        middle_data[count:num_middle_data]

                else:
                    sub_middle_data = middle_data[count:count + interval]

                new_data_lists.append(sub_middle_data)
                count += interval

        #count = 0

        #for rl in new_reports_lists:
        #    count += len(rl)

        return new_data_lists

    def add_arguments(self, parser):

        parser.add_argument('--bulk_create', 
            action = 'store_true', 
            dest = 'bulk_create', 
            default = False, 
            help = 'Stores new data as a bulk operation.  This option ' + \
                'will skip much of Django\'s save functionality, ' + \
                'but will provide faster speed for storing large numbers ' + \
                'of new data.')
        parser.add_argument('--previous_days', 
            type = int, 
            dest = 'previous_days', 
            default = 0, 
            help = 'Only ingest data from the given number of previous ' + \
                'days from when this command is run.')
        parser.add_argument('--update', 
            action = 'store_true', 
            dest = 'update', 
            default = False, 
            help = 'Update pre-existing data from the given TOA5 file.')
        parser.add_argument('--time_zone', 
            type = str, 
            dest = 'time_zone', 
            default = 'UTC', 
            help = 'The time zone to use for the timestamps in the ' + \
                'raw data file.')
        parser.add_argument('url', 
            type = str, 
            help = 'The remote or local URL to a CampbellSci TOA5 ' + \
                'formatted file.')

        return None

    def handle(self, *args, **options):

        # Get the command line options and required arguments, with 
        # default values.
        bulk_create = options.get('bulk_create', False)
        previous_days = options.get('previous_days', 0)
        update = options.get('update', False)
        time_zone = options.get('time_zone', 'UTC')
        url = options.get('url', '')

        ids = self.__parse_ids(url)
        station_id = None
        frequency_id = None

        if ids is not None:

            station_id, frequency_id = ids

            # We will attempt to parse the TOA5 file using the Pandas 
            # package.
            try:

                self.toa5 = pd.read_csv(urllib2.urlopen(url), 
                    index_col = 'TIMESTAMP', 
                    header = 1, 
                    skiprows = [2, 3], 
                    parse_dates = True, 
                    infer_datetime_format = True)
                #print toa5.columns

            except Exception as e:

                print str(e)
                self.toa5 = None

        else:
            self.toa5 = None

        if self.toa5 is not None:

            # Get the 'pdt' datetime, for use in determining if rows 
            # parsed when using the 'previous_days' option.
            pdt = now() - datetime.timedelta(days = previous_days)

            # Get the list of stations matching the parsed station_id
            self.stations = Station.objects.filter(station_id = station_id)

            # Create a UTC tzinfo object that will be used to replace 
            # the tzinfo attribute of parsed datetimes.  All ZiaMet 
            # timestamps must be in UTC already.
            tzinfo = pytz.timezone('UTC')

            first_dt = self.toa5.index.min().to_pydatetime()\
                .replace(tzinfo = tzinfo)
            latest_dt = self.toa5.index.max().to_pydatetime()\
                .replace(tzinfo = tzinfo)

            self.new_data = {}
            self.existing_data = {}
            self.model = FiveMinData

            if frequency_id == 'hourly':
                self.model = HourlyData

            _db_data = self.model.objects\
                .filter(rci = 0)\
                .filter(station__station_id = station_id)\
                .filter(dt__gte = first_dt)\
                .filter(dt__lte = latest_dt)

            for _d in _db_data:
                self.existing_data[str(_d.dt)] = _d

            # Get the number of rows in the dataframe.
            count = self.toa5.shape[0]

            for i in range(0, count):

                dt = self.toa5.iloc[i].name.to_pydatetime()\
                    .replace(tzinfo = tzinfo).astimezone(pytz.timezone('UTC'))
                row = self.toa5.iloc[i]

                parse = False

                if previous_days > 0:
                    if dt >= pdt:
                        parse = True

                else:
                    parse = True

                if parse is True:
                    #print dt
                    self.__parse_row(dt, row, bulk_create, update)

            if bulk_create is True:

                for dl in self.__split(self.new_data.values()):

                    try:
                        self.model.objects.bulk_create(dl)

                    except Exception as e:

                        err_str = 'Bulk Create Error: ' + str(e)
                        print err_str

        return None



