# Standard library imports.
import datetime

# PyTZ package imports.
import pytz

# Calculator module imports.
import dewpoint
import heat_index
import wind_chill
import gdd
import daily_etref

# Pandas package imports.
import pandas as pd

# Django package imports.
from django.db import models
from django.db.models.fields import FloatField


# These are the frequency ID values used to specify the frequency of 
# table data returned from the database using this module's included 
# managers.
FIVEMIN_FREQUENCY_ID = 'fivemin'
HOURLY_FREQUENCY_ID = 'hourly'
DAILY_FREQUENCY_ID = 'daily'

# The default timezone as a tzinfo object.
UTC = pytz.timezone('UTC')

# Here are the two types of indexes that may be used for the Pandas 
# dataframe.
DF_PERIOD_RANGE_INDEX = 'period_range'
DF_DATE_RANGE_INDEX = 'date_range'


def now():
    return UTC.localize(datetime.datetime.utcnow())


class BaseStationTimeseries(models.Manager):

    frequency_id = FIVEMIN_FREQUENCY_ID

    def __df_2_dict(self, df, tzinfo, **kwargs):

        # The return value of this method.  The return value is a 
        # dictionary containing a 'timeseries' key with the 
        # corresponding value as a list of dictionaries. The list 
        # elements correspond to the 'df' rows.
        tsdict = {'timeseries': df.to_dict(orient = 'records')}

        # A simpler reference to the timeseries list.
        ts = tsdict['timeseries']

        # Get the missing value option.  Create a list of types to 
        # check against.
        missing_value = kwargs.get('missing_value', None)
        types = [type(''), type(float('nan')), type(0)]

        # Variables used for looping over the dictionaries in the 'ts'
        # list.
        count = 0
        num_rows = len(ts)

        while count < num_rows:

            idx = df.index[count]
            row = ts[count]
            row['dt'] = datetime.datetime(idx.year, 
                idx.month, 
                idx.day, 
                idx.hour, 
                idx.minute, 
                idx.second, 
                tzinfo = tzinfo)

            for c in df.columns:

                row[c] = float(row[c])

                if type(missing_value) in types:
                    if str(row[c]) == str(float('nan')):
                        row[c] = missing_value

            count +=1

        return tsdict

    def _agg_dt(self, period, tzinfo):

        try:
            dt = datetime.datetime(period.year, 
                period.month, 
                period.day, 
                period.hour, 
                period.minute, 
                tzinfo = tzinfo)

        except:
            dt = None

        return dt

    def _aggregate(self, df, tzinfo):

        agg = {}

        for col in df.columns:

            if col.startswith('batt_volts') or \
                col.startswith('panel_temp') or \
                col.startswith('air_temp') or \
                col.startswith('rh') or \
                col.startswith('soil_temp'):

                agg[col] = {}
                agg[col]['max'] = df[col].max()
                agg[col]['max_dt'] = self._agg_dt(df[col].idxmax(), tzinfo)
                agg[col]['min'] = df[col].min()
                agg[col]['min_dt'] = self._agg_dt(df[col].idxmin(), tzinfo)
                agg[col]['mean'] = df[col].mean()

            if col.startswith('wind_speed'):

                agg[col] = {}
                agg[col]['max'] = df[col].max()
                agg[col]['max_dt'] = self._agg_dt(df[col].idxmax(), tzinfo)
                agg[col]['mean'] = df[col].mean()

            if col.startswith('wind_dir'):

                agg[col] = {}
                agg[col]['mean'] = df[col].mean()
                agg[col]['std'] = df[col].std()

            if col == 'precip_total':

                agg[col] = {}
                agg[col]['max'] = df[col].max()
                agg[col]['max_dt'] = self._agg_dt(df[col].idxmax(), tzinfo)
                agg[col]['sum'] = df[col].sum()

            if col in ('solar_rad', 'solar_rad_max', 'solar_rad_min', 
                'solar_rad_mean',):

                agg[col] = {}
                agg[col]['max'] = df[col].max()
                agg[col]['max_dt'] = self._agg_dt(df[col].idxmax(), tzinfo)
                agg[col]['min'] = df[col].min()
                agg[col]['min_dt'] = self._agg_dt(df[col].idxmin(), tzinfo)
                agg[col]['mean'] = df[col].mean()

            if col == 'solar_rad_total':

                agg[col] = {}
                agg[col]['max'] = df[col].max()
                agg[col]['max_dt'] = self._agg_dt(df[col].idxmax(), tzinfo)
                agg[col]['sum'] = df[col].sum()

        return agg

    def _to_imperial(self, df):

        for col in df.columns:

            if col.startswith('panel_temp') or \
                col.startswith('air_temp') or \
                col.startswith('soil_temp'):
                df[col] = (df[col] * (9.0 / 5.0)) + 32.0

            if col.startswith('wind_speed'):
                df[col] = df[col] * 2.237

            if col == 'precip_total':
                df[col] = df[col] / 25.4

        return None

    def as_df_objects(self, station_id, sdt, edt, **kwargs):

        # The 'df' variable is the return value of the method.  If 
        # there is no data in the database for the given start and 
        # end 'dts', the return value should be None.
        df = None

        # The 'objects' return value of this method.  Will return an 
        # None type if there is no data in the database.  Returns a 
        # dictionary of database objects, otherwise.
        objects = None

        tzinfo = sdt.tzinfo
        dfindex = kwargs.get('dfindex', DF_PERIOD_RANGE_INDEX)
        stnmeta = kwargs.get('stnmeta', False)

        # Retrieve the data for the station with the given 
        # 'station_id' attribute from the database.
        data = self.get_queryset()\
            .filter(rci = 0)\
            .filter(station__station_id = station_id)\
            .filter(dt__gte = sdt.astimezone(UTC))\
            .filter(dt__lte = edt.astimezone(UTC))\
            .select_related('station')\
            .order_by('dt')

        if data.count() > 0:

            sdt_str = sdt.strftime('%Y-%m-%d %H:%M')
            edt_str = edt.strftime('%Y-%m-%d %H:%M')
            freq = '5min'

            if self.frequency_id != FIVEMIN_FREQUENCY_ID:
                freq = 'H'

            if dfindex == DF_DATE_RANGE_INDEX:
                prng = pd.date_range(sdt_str, edt_str, freq = freq)

            else:
                prng = pd.period_range(sdt_str, edt_str, freq = freq)

            # Get the column names for the dataframe from the given 
            # models's float fields.  If 'stnmeta' is True, then add 
            # location columns from the related station object to the 
            # columns.
            columns = []
            ff_type = type(FloatField())

            for field in self.model._meta.get_fields():
                if type(field) == ff_type:
                    columns.append(field.name)

            if stnmeta is True:
                columns.extend(['latitude', 'longitude', 'elevation'])

            # Create the empty dataframe, setting the data columns to 
            # the float type.  Index is of course the period range 
            # created earlier.
            df = pd.DataFrame(index = prng, columns = columns, dtype = float)

            # Initialize the 'objects' return value to an empty 
            # dictionary.
            objects = {}

            # Iterate through the data objects, adding the values to 
            # the 'df' dataframe.  Is there a better way to make this 
            # dataframe?  Maybe answer that in the future.
            for d in data:

                dt = d.dt.astimezone(tzinfo)
                dt_str = dt.strftime('%Y-%m-%d %H:%M')

                try:
                    row = df.loc[dt_str]

                except:
                    row = None

                if row is not None:

                    objects[dt_str] = d

                    for c in columns:

                        if c not in ('latitude', 'longitude', 'elevation',):
                            row[c] = getattr(d, c, float('nan'))

                        else:
                            if stnmeta is True:
                                row[c] = getattr(d.station, c, float('nan'))

        return (df, objects)

    def as_df(self, station_id, sdt, edt, **kwargs):

        df, objects = self.as_df_objects(station_id, sdt, edt, **kwargs)

        return df

    def as_df_dict(self, station_id, sdt, edt, **kwargs):

        # Get the dataframe.
        df = self.as_df(station_id, sdt, edt, **kwargs)

        # Generate the dictionary of the 'df' dataframe, if 'df' is 
        # not None.  The 'tsdict' return value will also be None in 
        # that case.
        if df is not None:

            tsdict = self.__df_2_dict(df, sdt.tzinfo, **kwargs)

            if kwargs.get('aggregate', True) is True:
                tsdict['aggregates'] = self._aggregate(df, sdt.tzinfo)

        else:
            tsdict = None

        return (df, tsdict)

    def as_dict(self, station_id, sdt, edt, **kwargs):

        # Get the dataframe.
        df = self.as_df(station_id, sdt, edt, **kwargs)

        # Generate the dictionary of the 'df' dataframe, if 'df' is 
        # not None.  The 'tsdict' return value will be None in that 
        # case.
        if df is not None:

            tsdict = self.__df_2_dict(df, sdt.tzinfo, **kwargs)

            if kwargs.get('aggregate', False) is True:
                tsdict['aggregates'] = self._aggregate(df, sdt.tzinfo)

        else:
            tsdict = None

        return tsdict


class StationFiveMinTimeseries(BaseStationTimeseries):

    def _aggregate(self, df, tzinfo):

        agg = super(StationFiveMinTimeseries, self)._aggregate(df, tzinfo)

        for col in ('dewpoint', 'wind_chill', 'heat_index',):

            agg[col] = {}
            agg[col]['max'] = df[col].max()
            agg[col]['max_dt'] = self._agg_dt(df[col].idxmax(), tzinfo)
            agg[col]['min'] = df[col].min()
            agg[col]['min_dt'] = self._agg_dt(df[col].idxmin(), tzinfo)
            agg[col]['mean'] = df[col].mean()

        return agg

    def _to_imperial(self, df):

        super(StationFiveMinTimeseries, self)._to_imperial(df)

        for col in ('dewpoint', 'wind_chill', 'heat_index',):
            df[col] = (df[col] * (9.0 / 5.0)) + 32.0

        return None

    def as_df_objects(self, station_id, sdt, edt, **kwargs):

        # Adjust the start and end datetimes.  The start datetime is 
        # moved up to the nearest five minute interval, and the end 
        # datetime is moved down to the nearest five minute interval.
        if ((sdt.minute % 5) > 0) or (sdt.second > 0) or (sdt.microsecond > 0):

            sdt += datetime.timedelta(seconds = 300)
            sdt -= datetime.timedelta(minutes = sdt.minute % 5, 
                seconds = sdt.second, 
                microseconds = sdt.microsecond)

        edt -= datetime.timedelta(minutes = edt.minute % 5, 
            seconds = edt.second, 
            microseconds = edt.microsecond)

        # Get the dataframe.  The method may return a None value for 
        # the given parameters.
        df, objects = super(StationFiveMinTimeseries, self)\
            .as_df_objects(station_id, sdt, edt, **kwargs)

        if df is not None:

            # Calculate the dewpoint from the air temperature and the 
            # relative humidity.
            dp = lambda row: dewpoint.calculate(row['air_temp'], row['rh'])
            df['dewpoint'] = df.apply(dp, axis = 1)

            # Calculate the wind chill from the air temperature and the 
            # wind speed.
            wcc = lambda row: wind_chill.calculate(row['air_temp'], 
                row['wind_speed'] * 3.6)
            df['wind_chill'] = df.apply(wcc, axis = 1)

            # Calculate the heat index from the air temperature and the 
            # relative humidity.
            hi = lambda row: heat_index.calculate(row['air_temp'], row['rh'])
            df['heat_index'] = df.apply(hi, axis = 1)

            if kwargs.get('imperial', False) is True:
                self._to_imperial(df)

        return (df, objects)


class StationHourlyTimeseries(BaseStationTimeseries):

    frequency_id = HOURLY_FREQUENCY_ID

    def as_df_objects(self, station_id, sdt, edt, **kwargs):

        # Adjust the start and end datetimes.  The start datetime is 
        # moved up to the nearest hourly interval, and the end 
        # datetime is moved down to the nearest hourly interval.
        if (sdt.minute > 0) or (sdt.second > 0) or (sdt.microsecond > 0):

            sdt += datetime.timedelta(seconds = 3600)
            sdt -= datetime.timedelta(minutes = sdt.minute % 60, 
                seconds = sdt.second, 
                microseconds = sdt.microsecond)

        edt -= datetime.timedelta(minutes = edt.minute, 
            seconds = edt.second, 
            microseconds = edt.microsecond)

        # Get the dataframe.  The method may return a None value for 
        # the given parameters.
        df, objects = super(StationHourlyTimeseries, self)\
            .as_df_objects(station_id, sdt, edt, **kwargs)

        if df is not None:
            if kwargs.get('imperial', False) is True:
                self._to_imperial(df)

        return (df, objects)


class StationDailyTimeseries(BaseStationTimeseries):

    frequency_id = DAILY_FREQUENCY_ID

    def as_df_objects(self, station_id, sdt, edt, **kwargs):

        # Set the 'sdt' and 'edt' datetime parameters to the base days.
        sdt = datetime.datetime(sdt.year, 
            sdt.month, 
            sdt.day, 
            tzinfo = sdt.tzinfo)
        edt = datetime.datetime(edt.year,
            edt.month, 
            edt.day, 
            tzinfo = edt.tzinfo)

        # Adjust the 'sdt' and 'edt' datetime values to account for 
        # all the hours required to make a daily dataframe.
        adjsdt = (sdt + datetime.timedelta(seconds = 3600)).astimezone(UTC)
        adjedt = (edt + datetime.timedelta(days = 1)).astimezone(UTC)

        # Preserve the original value of the imperial key word 
        #argument.
        imperial = kwargs.get('imperial', False)

        # Set the imperial keyword argument to False, so imperial 
        # conversions are done in this method.
        kwargs['imperial'] = False

        # Sets the dfindex to 'period_range', as the 'date_range' does 
        # not work when reindexing to daily data.
        kwargs['dfindex'] = DF_PERIOD_RANGE_INDEX

        # The 'df' return value, set to None.
        df = None

        # Get the hourly dataframe.
        hdf, objects = super(StationDailyTimeseries, self)\
            .as_df_objects(station_id, adjsdt, adjedt, **kwargs)

        if hdf is not None:

            # Shifts the index by one hour into the past.
            hdf.index = hdf.index - 1

            r_dict = {}
            r_freq = 'D'

            for col in hdf.columns:

                if col.endswith('_max'):
                    r_dict[col] = hdf[col].resample(r_freq).max()

                if col.endswith('_min'):
                    r_dict[col] = hdf[col].resample(r_freq).min()

                if col.endswith('_mean'):
                    r_dict[col] = hdf[col].resample(r_freq).mean()

                if col.endswith('_total'):
                    r_dict[col] = hdf[col].resample(r_freq).sum()

                if col in ('latitude', 'longitude', 'elevation',):
                    r_dict[col] = hdf[col].resample(r_freq).mean()

            sdt_str = sdt.strftime('%Y-%m-%d')
            edt_str = edt.strftime('%Y-%m-%d')
            prng = pd.period_range(sdt_str, edt_str, freq = r_freq)
            df = pd.DataFrame(r_dict, index = prng)

            if imperial is True:
                self._to_imperial(df)

        return (df, objects)


class StationDailyRGTimeseries(StationDailyTimeseries):

    def _aggregate(self, df, tzinfo):

        agg = super(StationDailyRGTimeseries, self)._aggregate(df, tzinfo)

        agg['gdd'] = {}
        agg['gdd']['max'] = df['gdd'].max()
        agg['gdd']['max_dt'] = self._agg_dt(df['gdd'].idxmax(), tzinfo)
        agg['gdd']['mean'] = df['gdd'].mean()
        agg['gdd']['sum'] = df['gdd'].sum()

        for col in ('eth', 'eto', 'etr',):

            agg[col] = {}
            agg[col]['max'] = df[col].max()
            agg[col]['max_dt'] = self._agg_dt(df[col].idxmax(), tzinfo)
            agg[col]['min'] = df[col].min()
            agg[col]['min_dt'] = self._agg_dt(df[col].idxmin(), tzinfo)
            agg[col]['mean'] = df[col].mean()
            agg[col]['sum'] = df[col].sum()

        return agg

    def _to_imperial(self, df):

        super(StationDailyRGTimeseries, self)._to_imperial(df)

        for col in ('eth', 'eto', 'etr',):
            df[col] = df[col] / 25.4

        df['gdd'] = df['gdd'] * (9.0 / 5.0)

        return None

    def as_df_objects(self, station_id, sdt, edt, **kwargs):

        # Preserve the original value of the imperial keyword 
        # argument.
        imperial = kwargs.get('imperial', False)

        # Set the imperial keyword argument to False, so imperial 
        # conversions are done in this method.
        kwargs['imperial'] = False

        # Set the stnmeta keyword argument to True.  Doing so will 
        # populate the dataframe with location information.
        kwargs['stnmeta'] = True

        df, objects = super(StationDailyRGTimeseries, self)\
            .as_df_objects(station_id, sdt, edt, **kwargs)

        if df is not None:

            ts_f = lambda x: datetime.datetime(x.year, 
                x.month, 
                x.day, 
                x.hour, 
                x.minute, 
                x.second, 
                tzinfo = sdt.tzinfo)

            # Define the function that will calculate the reference ET 
            # values row-wise.
            et_f = lambda row: daily_etref\
                .calculate(ts_f(row.name).date(), 
                    row['latitude'],
                    row['elevation'], 
                    row['air_temp_max'], 
                    row['air_temp_min'], 
                    row['rh_max'], 
                    row['rh_min'], 
                    row['wind_speed_mean'], 
                    row['solar_rad_total'])

            # Create the columns for the ETref, then apply the et_f 
            # function row-wise to the dataframe.  I have tried to do 
            # the calculations in one step, but it was not working 
            # properly.
            df['eth'] = df.apply(lambda row: et_f(row)[0], axis = 1)
            df['eto'] = df.apply(lambda row: et_f(row)[1], axis = 1)
            df['etr'] = df.apply(lambda row: et_f(row)[2], axis = 1)

            # Define the function that will calculate the growing 
            # degree days row-wise.
            #gdd_f = lambda row: gdd.calculate(row['air_temp_max'], 
            #    row['air_temp_min'], **kwargs)

            # Define the function that will calculate the cumulative 
            # degree days column-wise.
            cgdd_offset = kwargs.get('cgdd_offset', 10.0)
            cgdd_f = lambda x: cgdd_offset + x.cumsum()

            # Calculate the growing degree days.  Wait until the values 
            # are converted to imperial (if necessary) to calculate the 
            # cumulative gdd.
            localgdd = lambda row: gdd.calculate(row['air_temp_max'], 
                row['air_temp_min'], **kwargs)
            df['gdd'] = df.apply(localgdd, axis = 1)
            df['cgdd'] = float('nan')

            if imperial is True:
                self._to_imperial(df)

            # Calculate the cumulative sum of the growing degree days, 
            # using a year column to group the data.
            df['year'] = [t.year for t in df.index]
            df['cgdd'] = df.groupby('year')['gdd'].apply(cgdd_f)

        return (df, objects)


