# Standard library imports.
import datetime
import uuid

# PyTZ package imports.
import pytz

# Django package imports.
from django.db import models

# Local package imports.
from .managers import UTC
from .managers import now
from .managers import StationFiveMinTimeseries
from .managers import StationHourlyTimeseries
from .managers import StationDailyTimeseries
from .managers import StationDailyRGTimeseries


# These are the default start and end datetime for the Station 
# classes.
DEFAULT_SDT = UTC.localize(datetime.datetime(1900, 1, 1, 0, 0, 0))
DEFAULT_EDT = UTC.localize(datetime.datetime(9999, 12, 31, 23, 59, 59))


# #####################################################################
# Application model implementations begin here.
# #####################################################################

class BaseModel(models.Model):
    """The abstract base class for the ZiaMet models.

    Attributes
    ----------
    _id : uuid
        This is the primary key field used for database storage.  
    created : datetime
        A UTC datetime that denotes when the model information was 
        stored in the database.
    last_modified : datetime
        A UTC datetime that denotes when the model information was 
        updated in the database.
    description : string
        Stores a description of the model information.  Information 
        stored in this attribute could be anything about the model 
        that is not easily quantifiable.
    """

    _id = models.UUIDField(primary_key = True, default = uuid.uuid4, 
        editable = False)
    created = models.DateTimeField(auto_now_add = True)
    last_modified = models.DateTimeField(auto_now = True)
    description = models.TextField(blank = True, null = True)

    class Meta:
        abstract = True


class Station(BaseModel):
    """Represents a surface weather station.

    This model is used to describe a surface weather station that 
    observes a standard set of variables.

    Attributes
    ----------
    station_id : string
        The public identifier for the Station.
    sdt : datetime
        A UTC datetime that denotes when the Station was deployed.
    edt : datetime
        A UTC datetime that denotes when the Station was discontinued. 
        The default value is the last second of the year 9999.
    name : string
        The name of the Station.
    latitude : float
        The latitude location of the Station, as a decimal number.
    longitude : float
        The longitude location of the Station, as a decimal number.
    elevation : float
        The elevation of the Station, in meters, as a decimal number.
    time_zone : string
        Represents the time zone where the Station is located.

    """
    station_id = models.CharField(max_length = 32, verbose_name = 'Station ID')
    sdt = models.DateTimeField(default = now)
    edt = models.DateTimeField(default = DEFAULT_EDT)
    name = models.CharField(max_length = 256)
    latitude = models.FloatField(default = 32.282)
    longitude = models.FloatField(default = -106.760)
    elevation = models.FloatField(default = 1186.0)
    time_zone = models.CharField(max_length = 64, verbose_name = 'Time Zone', 
        blank = False, null = False, default = 'US/Mountain')

    class Meta:
        unique_together = ('station_id', 'sdt')

    # #################################################################
    # The following are the Python "magic" methods.
    # #################################################################

    def __unicode__(self):
        """Returns the unicode representation of the Station.

        Returns
        -------
        unicode
            A unicode represnetation of the Station, including the 
            'name', 'station_id', and 'sdt' attributes.

        """
        name = self.name
        date_str = unicode(self.sdt.date())

        if name is None or name == '':
            name = self.station_id

        u_tmpl = '{name} ({station_id}, {date_str})'
        u = u_tmpl.format(name = name, 
            station_id = self.station_id, 
            date_str = date_str)

        return u

    # #################################################################
    # The following methods belong to an instance of the Station 
    # model class.
    # #################################################################

    def elevation_feet(self):
        """A method that returns the station's elevation in feet.

        Returns
        -------
        float
            The value of the station's 'elevation' attribute, in feet.

        """
        return self.elevation * 3.281

    def get_tzinfo(self):
        """Return a tzinfo object representing the station's time zone.

        Returns
        -------
        tzinfo
            A tzinfo object.

        """
        return pytz.timezone(self.time_zone)

    def valid_dt(self, dt):
        """Determines if a 'dt' is valid for the Station.

        The 'dt' is checked agains the 'sdt' and 'edt' attributes to 
        determine if the 'dt' parameter falls within the datetime range 
        of the 'sdt' and 'edt' attribute values.

        Parameters
        ----------
        dt : datetime
            A UTC datetime.

        Returns
        -------
        boolean
            True, if the given 'dt' parameter falls within the 'sdt' 
            and 'edt' attribute values, False otherwise.
        """
        is_valid = False

        try:
            if (dt >= self.sdt) and (dt < self.edt):
                is_valid = True

        except:
            pass

        return is_valid


class BaseData(BaseModel):

    # #################################################################
    # The following manager assignment is to preserve the objects 
    # manager for the child classes.  See:
    # https://docs.djangoproject.com/en/1.8/topics/db/managers
    # /#default-managers
    # #################################################################

    objects = models.Manager()

    # #################################################################
    # The following define the database fields for the base model.
    # #################################################################

    station = models.ForeignKey('Station')
    dt = models.DateTimeField(blank = False, null = False)
    rci = models.PositiveIntegerField(default = 0)

    class Meta:

        abstract = True
        unique_together = ('station', 'dt', 'rci')

    # #################################################################
    # The following methods are extended/overridden from the base 
    # Model class.
    # #################################################################

    def save(self, *args, **kwargs):

        if self.station.valid_dt(self.dt) is True:
            super(BaseData, self).save(*args, **kwargs)

        else:

            err_str = 'The \'dt\' attribute is not within the active ' + \
                'datetime range of the \'station\' attribute, specified ' + \
                'by the \'sdt\' and \'edt\' \'station\' attributes.'

            raise Exception(err_str)

        return None


class FiveMinData(BaseData):

    # #################################################################
    # The following are custom managers for this model.
    # #################################################################

    station_timeseries = StationFiveMinTimeseries()

    # #################################################################
    # The following define the database fields for the model.
    # #################################################################

    record = models.PositiveIntegerField(blank = True, null = True)
    batt_volts = models.FloatField(blank = True, null = True)
    panel_temp = models.FloatField(blank = True, null = True)
    air_temp = models.FloatField(blank = True, null = True)
    rh = models.FloatField(blank = True, null = True)
    wind_speed = models.FloatField(blank = True, null = True)
    wind_speed_max = models.FloatField(blank = True, null = True)
    wind_speed_mean = models.FloatField(blank = True, null = True)
    wind_dir = models.FloatField(blank = True, null = True)
    wind_dir_mean = models.FloatField(blank = True, null = True)
    precip_total = models.FloatField(blank = True, null = True)
    solar_rad = models.FloatField(blank = True, null = True)
    solar_rad_total = models.FloatField(blank = True, null = True)
    soil_temp = models.FloatField(blank = True, null = True)

    # #################################################################
    # The following are the Python "magic" methods.
    # #################################################################

    def __unicode__(self):
        """Returns the unicode representation of the FiveMinData.

        Returns
        -------
        unicode
            The unicode representation of the FiveMinData, which 
            includes the 'dt' and 'station.name' attributes.

        """
        u_tmpl = '{dt} {station_name} Five Minute Data'
        station_name = self.station.name

        if station_name is None or station_name == '':
            station_name = self.station.station_id

        u = u_tmpl.format(dt = str(self.dt), 
            station_name = station_name)

        return u


class HourlyData(BaseData):

    # #################################################################
    # The following are custom managers for this model.
    # #################################################################

    station_timeseries = StationHourlyTimeseries()
    station_daily_timeseries = StationDailyTimeseries()
    station_daily_rg_timeseries = StationDailyRGTimeseries()

    # #################################################################
    # The following define the database fields for the model.
    # #################################################################

    record = models.PositiveIntegerField(blank = True, null = True)
    batt_volts_max = models.FloatField(blank = True, null = True)
    batt_volts_min = models.FloatField(blank = True, null = True)
    batt_volts_mean = models.FloatField(blank = True, null = True)
    panel_temp_max = models.FloatField(blank = True, null = True)
    panel_temp_min = models.FloatField(blank = True, null = True)
    panel_temp_mean = models.FloatField(blank = True, null = True)
    air_temp_max = models.FloatField(blank = True, null = True)
    air_temp_min = models.FloatField(blank = True, null = True)
    air_temp_mean = models.FloatField(blank = True, null = True)
    rh_max = models.FloatField(blank = True, null = True)
    rh_min = models.FloatField(blank = True, null = True)
    rh_mean = models.FloatField(blank = True, null = True)
    wind_speed_max = models.FloatField(blank = True, null = True)
    wind_speed_mean = models.FloatField(blank = True, null = True)
    wind_dir_mean = models.FloatField(blank = True, null = True)
    precip_total = models.FloatField(blank = True, null = True)
    solar_rad_max = models.FloatField(blank = True, null = True)
    solar_rad_min = models.FloatField(blank = True, null = True)
    solar_rad_mean = models.FloatField(blank = True, null = True)
    solar_rad_total = models.FloatField(blank = True, null = True)
    soil_temp_max = models.FloatField(blank = True, null = True)
    soil_temp_min = models.FloatField(blank = True, null = True)
    soil_temp_mean = models.FloatField(blank = True, null = True)

    # #################################################################
    # The following are the Python "magic" methods.
    # #################################################################

    def __unicode__(self):
        """Returns the unicode representation of the HourlyData.

        Returns
        -------
        unicode
            The unicode representation of the HourlyData, which 
            includes the 'dt' and 'station.name' attributes.

        """
        u_tmpl = '{dt} {station_name} Hourly Data'
        station_name = self.station.name

        if station_name is None or station_name == '':
            station_name = self.station.station_id

        u = u_tmpl.format(dt = str(self.dt), 
            station_name = station_name)

        return u


