# Standard library imports.
import datetime

# PyTZ package imports.
import pytz

# Django package imports.
from django.db.models import Max
from django.http import HttpResponse
from django.http import HttpResponseBadRequest
from django.template import loader
from django.template import RequestContext
from django.views.generic import View

# Local package imports.
from .forms import FIVEMIN_FREQUENCY_ID
from .forms import HOURLY_FREQUENCY_ID
from .forms import DAILY_FREQUENCY_ID
from .forms import GRAPHICAL_OUTPUT
from .forms import TABLE_OUTPUT
from .forms import CSV_OUTPUT
from .forms import IMPERIAL_UNITS
from .forms import StationTimeseries as StnTimeseriesF
from .forms import StationTimeseriesPrev as StnTimeseriesPrevF
from .forms import StationDailyRGTimeseries as StnDailyRGTimeseriesF
from .forms import StationDailyRGTimeseriesPrev as StnDailyRGTimeseriesPrevF
from .models import UTC
from .models import now
from .models import Station
from .models import FiveMinData
from .models import HourlyData


class Index(View):

    def get(self, request, *args, **kwargs):

        # Get the ZiaMet stations with the max dt from the hourly data.
        stations = Station.objects.filter(edt__gte = now())\
            .order_by('name')\
            .annotate(max_dt = Max('hourlydata__dt'))

        for station in stations:

            ltz = pytz.timezone(station.time_zone)
            station.max_dt = station.max_dt.astimezone(ltz)

        context = {}
        context['stations'] = stations

        tp = 'nmcc_ziamet/index.html'
        t = loader.get_template(tp)
        c = RequestContext(request, context)

        return HttpResponse(t.render(c))


class Status(View):

    def __status(self, station):

        self.__status_latest_report(station)

        return None

    def __status_latest_report(self, station):

        # Create some threshold datetimes to check the latest 
        # timestamp against.
        n = now()
        caution = n - datetime.timedelta(seconds = 3600)
        minor = n - datetime.timedelta(seconds = 3600 * 3)
        major = n - datetime.timedelta(seconds = 3600 * 6)
        severe = n - datetime.timedelta(seconds = 3600 * 12)

        if station.max_dt is not None:

            if station.max_dt <= caution and \
                station.max_dt > minor:

                issue = 'Last report more than an hour ago.'
                station.status_severity_index += 1
                station.status_issues_list.append(issue)

            elif station.max_dt <= minor and \
                station.max_dt > major:

                issue = 'Last report more than three hours ago.'
                station.status_severity_index += 2
                station.status_issues_list.append(issue)

            elif station.max_dt <= major and \
                station.max_dt > severe:

                issue = 'Last report more than six hours ago.'
                station.status_severity_index += 3
                station.status_issues_list.append(issue)

            elif station.max_dt <= severe:

                issue = 'Last report more than 12 hours ago. ' + \
                    'Resolve immediately.'
                station.status_severity_index += 4
                station.status_issues_list.append(issue)

        else:

            issue = 'Data does not exist for this station. Ingest data as ' + \
                'soon as possible.'
            station.status_severity_index += 3
            station.status_issues_list.append(issue)

        return None

    def get(self, request, *args, **kwargs):

        # Get the ZiaMet stations with the max dt from the five minute 
        # data.
        stations = []
        stations_qs = Station.objects.filter(edt__gte = now())\
            .order_by('name')\
            .annotate(max_dt = Max('fivemindata__dt'))

        for station in stations_qs:

            station.status_severity_index = 0
            station.status_issues_list = []
            self.__status(station)
            stations.append(station)

        # Sort the stations according to the severity index.
        stations.sort(key = lambda x: x.status_severity_index)
        stations.reverse()

        context = {}
        context['stations'] = stations

        tp = 'nmcc_ziamet/status.html'
        t = loader.get_template(tp)
        c = RequestContext(request, context)

        return HttpResponse(t.render(c))


class StationIndex(View):

    def __since_midnight(self, station, df, edt):

        # Trim the dataframe to start from the local midnight.
        ltz = edt.tzinfo
        lm = ltz.localize(datetime.datetime(edt.year, edt.month, edt.day))
        lm_str = lm.strftime('%Y-%m-%d 00:00')
        lmdf = df[lm_str:]

        # This is a bit of a trick.  The '_aggregate' method is not 
        # meant to be used as a 'public' method, but it can still be 
        # accessed externally.  The aggregates of the trimmed dataframe 
        # are returned using this method.
        agg = FiveMinData.station_timeseries._aggregate(df, ltz)

        return agg

    def get(self, request, *args, **kwargs):

        station_id = kwargs.get('station_id', None)
        station = None

        if station_id is not None:

            stations = Station.objects.filter(edt__gte = now())\
                .filter(station_id = station_id)\
                .annotate(max_dt = Max('fivemindata__dt'))

            if stations.count() == 1:
                station = stations[0]

        # Create the context dict here, so that if the station doesn't 
        # exist, we don't have to do checking at the end of this 
        # method.
        context = {}
        context['station_id'] = station_id
        context['station'] = station

        if station is not None:

            delta = datetime.timedelta(seconds = 3600 * 6)

            if station.max_dt >= (now() - delta):

                tzinfo = station.get_tzinfo()
                sdt = station.max_dt - datetime.timedelta(days = 1)
                sdt += datetime.timedelta(seconds = 300)
                sdt = sdt.astimezone(tzinfo)

                edt = station.max_dt
                edt = edt.astimezone(tzinfo)

                df, tsdict = FiveMinData.station_timeseries\
                    .as_df_dict(station_id, sdt, edt, 
                        imperial = True, 
                        missing_value = 0.0)

                current = tsdict['timeseries'][-1]
                current['precip_total'] = df['precip_total'].sum()

                context['current'] = current
                context['timeseries'] = tsdict['timeseries']                
                context['since_midnight'] = self\
                    .__since_midnight(station, df, edt)

            else:
                setattr(station, 'inactive', True)

        tp = 'nmcc_ziamet/station/index.html'
        t = loader.get_template(tp)
        c = RequestContext(request, context)

        return HttpResponse(t.render(c))


class StationTimeseries(View):

    def get_station(self, station_id):

        station = None

        stations = Station.objects.filter(station_id = station_id)\
            .annotate(max_dt = Max('hourlydata__dt'))\
            .order_by('sdt')
        count = stations.count()

        if count >= 1:
            station = stations[count - 1]

        return station

    def get(self, request, *args, **kwargs):

        station_id = kwargs.get('station_id', None)
        station = self.get_station(station_id)

        context = {}
        context['station_id'] = station_id
        context['station'] = station

        # The default template and mimetype output to use for the HTTP 
        # response.
        template = 'nmcc_ziamet/station/timeseries.html'
        mimetype = 'text/html'

        if station is not None:

            form = StnTimeseriesF(kwargs)

            if form.is_valid():

                # Populate the response context with some important 
                # parameter values that will affect template rendering.
                context['frequency_id'] = form.cleaned_data['frequency_id']
                context['time_zone'] = form.cleaned_data['time_zone']
                context['output'] = form.cleaned_data['output']
                context['imperial'] = form.cleaned_data['imperial']

                # Determine the function to use to retrieve data from 
                # the database based on the 'frequency_id' parameter 
                # value.
                if form.cleaned_data['frequency_id'] == FIVEMIN_FREQUENCY_ID:
                    tsm = FiveMinData.station_timeseries.as_dict

                elif form.cleaned_data['frequency_id'] == HOURLY_FREQUENCY_ID:
                    tsm = HourlyData.station_timeseries.as_dict

                else:
                    tsm = HourlyData.station_daily_timeseries.as_dict

                # In this block, we will determine the template and 
                # mimetype for the response, as well as the value to 
                # use when timeseries values are missing, based on the 
                # requested output.
                if form.cleaned_data['output'] == GRAPHICAL_OUTPUT:
                    missing_value = 0.0

                elif form.cleaned_data['output'] == CSV_OUTPUT:

                    missing_value = float('nan')
                    template = 'nmcc_ziamet/station/hd_timeseries.csv'
                    mimetype = 'text/csv'

                    if form.cleaned_data['frequency_id'] == \
                        FIVEMIN_FREQUENCY_ID:
                        template = 'nmcc_ziamet/station/fivemin_timeseries.csv'

                else:
                    missing_value = ''

                # Get the timeseries using the method determined in a
                # prior if/then block.
                ts = tsm(station_id, 
                    form.cleaned_data['sdt'], 
                    form.cleaned_data['edt'], 
                    imperial = form.cleaned_data['imperial'], 
                    missing_value = missing_value)

                if ts is not None:
                    context['timeseries'] = ts['timeseries']

                else:

                    context['no_data'] = True
                    context['form'] = form

            else:
                context['form'] = form

        t = loader.get_template(template)
        c = RequestContext(request, context)

        return HttpResponse(t.render(c), content_type = mimetype)


class StationTimeseriesForToday(StationTimeseries):

    def get(self, request, *args, **kwargs):

        station_id = kwargs.get('station_id', None)
        station = self.get_station(station_id)

        if station is not None:

            ltz = pytz.timezone(station.time_zone)
            today = now().astimezone(ltz)
            today = today.date()

            #kwargs['station_id'] = station_id
            kwargs['frequency_id'] = HOURLY_FREQUENCY_ID
            kwargs['start_date'] = str(today)
            kwargs['end_date'] = str(today)
            kwargs['time_zone'] = station.time_zone
            kwargs['units'] = IMPERIAL_UNITS
            kwargs['output'] = TABLE_OUTPUT

        r = super(StationTimeseriesForToday, self)\
            .get(request, *args, **kwargs)

        return r


class StationTimeseriesPrev(StationTimeseries):

    def get(self, request, *args, **kwargs):

        form = StnTimeseriesPrevF(kwargs)

        if form.is_valid():

            kwargs.update(form.cleaned_data)

            r = super(StationTimeseriesPrev, self)\
                .get(request, *args, **kwargs)

        else:
            r = HttpResponseBadRequest('Bad Request!')

        return r


class StationRequestTimeseries(StationTimeseries):

    def get(self, request, *args, **kwargs):

        station_id = kwargs.get('station_id', None)
        station = self.get_station(station_id)

        context = {}
        context['station_id'] = station_id
        context['station'] = station

        if station is not None:

            tzinfo = pytz.timezone(station.time_zone)
            st = (station.max_dt - datetime.timedelta(days = 31))\
                .astimezone(tzinfo)
            et = station.max_dt.astimezone(tzinfo)

            sd = st.date()
            ed = et.date()
            initial = {'start_date': sd, 
                'end_date': ed, 
                'time_zone': station.time_zone}

            context['form'] = StnTimeseriesF(initial = initial)

        t = loader.get_template('nmcc_ziamet/station/timeseries.html')
        c = RequestContext(request, context)

        return HttpResponse(t.render(c))

    def post(self, request, *args, **kwargs):

        kwargs['station_id'] = request.POST.get('station_id')
        kwargs['frequency_id'] = request.POST.get('frequency_id')
        kwargs['start_date'] = request.POST.get('start_date')
        kwargs['end_date'] = request.POST.get('end_date')
        kwargs['time_zone'] = request.POST.get('time_zone')
        kwargs['units'] = request.POST.get('units')
        kwargs['output'] = request.POST.get('output')

        r = super(StationRequestTimeseries, self).get(request, *args, **kwargs)

        return r


class StationRGTimeseries(StationTimeseries):

    def get(self, request, *args, **kwargs):

        # The frequency for a Reference ET and GDD request will ALWAYS 
        # be 'daily'.
        kwargs['frequency_id'] = DAILY_FREQUENCY_ID

        station_id = kwargs.get('station_id', None)
        station = self.get_station(station_id)

        context = {}
        context['station_id'] = station_id
        context['station'] = station

        template = 'nmcc_ziamet/station/retgdd_timeseries.html'
        mimetype = 'text/html'

        if station is not None:

            form = StnDailyRGTimeseriesF(kwargs)

            if form.is_valid():

                context['frequency_id'] = form.cleaned_data['frequency_id']
                context['time_zone'] = form.cleaned_data['time_zone']
                context['output'] = form.cleaned_data['output']
                context['imperial'] = form.cleaned_data['imperial']

                if context['output'] == GRAPHICAL_OUTPUT:
                    missing_value = 0.0

                elif context['output'] == CSV_OUTPUT:

                    missing_value = float('nan')
                    template = 'nmcc_ziamet/station/retgdd_timeseries.csv'
                    mimetype = 'text/csv'

                else:
                    missing_value = ''

                ts = HourlyData.station_daily_rg_timeseries.as_dict(
                    station_id, 
                    form.cleaned_data['sdt'], 
                    form.cleaned_data['edt'], 
                    imperial = form.cleaned_data['imperial'], 
                    missing_value = missing_value, 
                    gdd_tbase = form.cleaned_data['gdd_tbase'], 
                    gdd_tmaxcutoff = form.cleaned_data['gdd_tmaxcutoff'], 
                    gdd_tmincutoff = form.cleaned_data['gdd_tmincutoff'], 
                    cgdd_offset = form.cleaned_data['cgdd_offset'])

                if ts is not None:
                    context['timeseries'] = ts['timeseries']

                else:

                    context['no_data'] = True
                    context['form'] = form

            else:
                context['form'] = form

        t = loader.get_template(template)
        c = RequestContext(request, context)

        return HttpResponse(t.render(c), content_type = mimetype)


class StationRGTimeseriesPrev(StationRGTimeseries):

    def get(self, request, *args, **kwargs):

        # The frequency for a Reference ET and GDD request will ALWAYS 
        # be 'daily'.
        kwargs['frequency_id'] = DAILY_FREQUENCY_ID

        form = StnDailyRGTimeseriesPrevF(kwargs)

        if form.is_valid():

            kwargs.update(form.cleaned_data)

            r = super(StationRGTimeseriesPrev, self)\
                .get(request, *args, **kwargs)

        else:

            print form.errors
            r = HttpResponseBadRequest('Bad Request!')

        return r


class StationRequestRGTimeseries(StationRGTimeseries):

    def get(self, request, *args, **kwargs):

        station_id = kwargs.get('station_id', None)
        station = self.get_station(station_id)

        context = {}
        context['station_id'] = station_id
        context['station'] = station

        if station is not None:

            tzinfo = pytz.timezone(station.time_zone)
            st = (station.max_dt - datetime.timedelta(days = 31))\
                .astimezone(tzinfo)
            et = station.max_dt.astimezone(tzinfo)

            sd = st.date()
            ed = et.date()
            initial = {'start_date': sd, 
                'end_date': ed, 
                'time_zone': station.time_zone}

            context['form'] = StnDailyRGTimeseriesF(initial = initial)

        t = loader.get_template('nmcc_ziamet/station/retgdd_timeseries.html')
        c = RequestContext(request, context)

        return HttpResponse(t.render(c))

    def post(self, request, *args, **kwargs):

        kwargs['station_id'] = request.POST.get('station_id')
        kwargs['frequency_id'] = request.POST.get('frequency_id')
        kwargs['start_date'] = request.POST.get('start_date')
        kwargs['end_date'] = request.POST.get('end_date')
        kwargs['time_zone'] = request.POST.get('time_zone')
        kwargs['units'] = request.POST.get('units')
        kwargs['output'] = request.POST.get('output')
        kwargs['gdd_tbase'] = request.POST.get('gdd_tbase')
        kwargs['gdd_tmaxcutoff'] = request.POST.get('gdd_tmaxcutoff')
        kwargs['gdd_tmincutoff'] = request.POST.get('gdd_tmincutoff')
        kwargs['cgdd_offset'] = request.POST.get('cgdd_offset')

        r = super(StationRequestRGTimeseries, self)\
            .get(request, *args, **kwargs)

        return r


